'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global Fennec, logging, STATS, log, translate, browser, portInit
parseUri, ElementTypes, getSettings, DataCollection, loggingEnable,
Settings, storageSet, storageGet, MyFilters, getUnicodeUrl, FilterNormalizer,
Filter, PatternFilter, getUnicodeDomain, determineUserLanguage */

/* eslint-disable no-use-before-define */

let settingsObj = {};
let myfiltersObj = {};

const initializeBackgroundObjects = function () {
  if (Fennec) {
    browser.browserAction.onClicked.addListener((tab) => {
      openTab(`button/popup.html?tabId=${tab.id}&mobile=1`);
    });
    browser.runtime.onMessage.addListener((request, sender, sendResponse) => {
      if (!(request.message === 'closePopupTab')) {
        return;
      }
      if (sender && sender.tab && sender.tab.id) {
        browser.tabs.remove(sender.tab.id);
      }
      sendResponse({});
    });
  }

  settingsObj = new Settings();
  settingsObj.onload().then(() => {
    if (getSettings().debug_logging) {
      logging(true);
    }
  });

  if (typeof STATS === 'object') {
    STATS.startPinging();
  }

  if (typeof blockCounts === 'object') {
    blockCounts.init();

    if (browser.runtime.setUninstallURL) {
      STATS.checkUserId().then((userID) => {
        const uninstallURL = `https://getadblock.com/uninstall/?u=${userID}`;
        // if the start property of blockCount exists (which is the AdBlock installation timestamp)
        // use it to calculate the approximate length of time that user has AdBlock installed
        if (blockCounts && typeof blockCounts.get === 'function') {
          const twoMinutes = 2 * 60 * 1000;
          const updateUninstallURL = function () {
            const localBC = blockCounts.get();
            if (localBC && localBC.start) {
              const installedDuration = (Date.now() - localBC.start);
              let url = `${uninstallURL}&t=${installedDuration}`;
              const bc = localBC.total;
              url = `${url}&bc=${bc}`;
              if (typeof myfiltersObj === 'object' &&
                  myfiltersObj.subscriptions &&
                  myfiltersObj.subscriptions.adblock_custom &&
                  myfiltersObj.subscriptions.adblock_custom.lastUpdate) {
                url = `${url}&abc-lt=${myfiltersObj.subscriptions.adblock_custom.lastUpdate}`;
              } else {
                url += '&abc-lt=-1';
              }

              browser.runtime.setUninstallURL(url);
            }
          };
          // start an interval timer that will update the Uninstall URL every 2 minutes
          setInterval(updateUninstallURL, twoMinutes);
          updateUninstallURL();
        } else {
          browser.runtime.setUninstallURL(`${uninstallURL}&t=-1`);
        }
      });
    }
  }

  countCacheInit();
  myfiltersObj = new MyFilters((() => {
    myfiltersObj.init();
  })); // end of new MyFilters
  myfiltersObj.ready().then(() => {
    browser.tabs.query({ url: 'http://*/*' }, handleEarlyOpenedTabs);
    browser.tabs.query({ url: 'https://*/*' }, handleEarlyOpenedTabs);
    browser.webRequest.onBeforeRequest.addListener(onBeforeRequestHandler, { urls: ['http://*/*', 'https://*/*'] }, ['blocking']);
    browser.tabs.onRemoved.addListener(frameData.removeTabId);
    // Popup blocking
    if (browser.webNavigation && browser.webNavigation.onCreatedNavigationTarget) {
      browser.webNavigation.onCreatedNavigationTarget.addListener(onCreatedNavigationTargetHandler);
    }
  });

  const openInstalledTab = function () {
    const installedURL = `https://getadblock.com/installed/?u=${STATS.userId()}`;
    browser.tabs.create({ url: installedURL });
  };

  // Open the install tab when firstRun is true
  STATS.checkUserId().then(() => {
    if (STATS.firstRun) {
      openInstalledTab();
    }
  });
}; // end of initializeBackgroundObjects

// Records how many ads have been blocked by AdBlock.  This is used
// by the AdBlock app in the Chrome Web Store to display statistics
// to the user.
// Block Counts are stored in 'localstorage' because of performance requirements
// in the recordOneAdBlocked() function.  Data loss can occur when using storage.local.get/set
// in quick succession.
// In Firefox v54, a user clearing all history, also clears localStorage for web extensions,
// so a periodic backup of the block counts is stored in browser.storage.
// If an empty object is returned from localStorage, the backup is retrieved and save.
// This may cause a minor loss of total block counts depending on the timing of saving of
// the backup, and the clearing of history.
const blockCounts = (function blockCountsFunction() {
  const BCkey = 'blockage_stats';
  const fiveMinutes = 300000;

  const reloadFromBackup = function () {
    browser.storage.local.get('backup_blockcount', (response) => {
      const backupBC = response.backup_blockcount;
      if (backupBC) {
        storageSet(BCkey, backupBC);
      }
    });
  };

  browser.runtime.onMessage.addListener((request, sender, sendResponse) => {
    if (!(request.message === 'recordOneAdBlocked')) {
      return;
    }
    if (sender && sender.tab && sender.tab.id) {
      blockCounts.recordOneAdBlocked(sender.tab.id);
      updateBadge(sender.tab.id);
    }
    sendResponse({});
  });

  window.setInterval(() => {
    const bc = blockCounts.get();
    browser.storage.local.get('backup_blockcount', (response) => {
      const backupBC = response.backup_blockcount;
      if (!backupBC && bc) {
        browser.storage.local.set({ backup_blockcount: bc });
        return;
      }
      if (bc && backupBC && bc.total > backupBC.total) {
        browser.storage.local.set({ backup_blockcount: bc });
      }
    });
  }, fiveMinutes);

  return {
    init() {
      let data = storageGet(BCkey);
      if (!data) {
        data = {};
      }
      if (data.start === undefined) {
        data.start = Date.now();
      }
      if (data.total === undefined) {
        data.total = 0;
      }
      if (data.malware_total === undefined) {
        data.malware_total = 0;
      }
      data.version = 1;
      storageSet(BCkey, data);
    },

    recordOneAdBlocked(tabId) {
      const data = storageGet(BCkey);
      if (!data) {
        reloadFromBackup();
        return;
      }
      data.total += 1;
      storageSet(BCkey, data);

      // code for incrementing ad blocks
      const currentTab = frameData.get(tabId);
      if (currentTab) {
        currentTab.blockCount += 1;
      }
    },

    recordOneMalwareBlocked() {
      const data = storageGet(BCkey);
      if (!data) {
        reloadFromBackup();
        return;
      }
      data.malware_total += 1;
      storageSet(BCkey, data);
    },

    get() {
      let data = storageGet(BCkey);
      // since we need to return 'something', re-init, then restore from the last backup
      // the restore is done async, so it should be an immediate impact.
      if (!data) {
        blockCounts.init();
        data = storageGet(BCkey);
        reloadFromBackup();
      }
      return data;
    },

    getTotalAdsBlocked(tabId) {
      if (tabId) {
        const currentTab = frameData.get(tabId);
        return currentTab ? currentTab.blockCount : 0;
      }

      return this.get().total;
    },
  };
}());


// called from bandaids, for use on our getadblock.com site
const getAdblockUserId = function (callbackFN) {
  if (!callbackFN && (typeof callbackFN !== 'function')) {
    return;
  }

  browser.storage.local.get(STATS.userIDStorageKey, (response) => {
    callbackFN(response[STATS.userIDStorageKey]);
  });
};

// called from bandaids, for use on our getadblock.com site
const getFirstRun = function () {
  return STATS.firstRun;
};

// called from bandaids, for use on our getadblock.com site
const setFirstRunToFalse = function () {
  STATS.firstRun = false;
};

// Open a new tab with a given URL.
// Inputs:
//   url: string - url for the tab
//   nearActive: bool - open the tab near currently active (instead of at the end).
//               optional, defaults to false
function openTab(url, nearActive, callback) {
  // create an empty callback function, if we were called without on to
  // prevent JS errors.
  let aCallback = callback;
  if (typeof aCallback !== 'function') {
    aCallback = function () {
    };
  }
  if (Fennec) {
    browser.tabs.create({ url, active: true }, aCallback);
    return;
  }
  browser.windows.getCurrent((current) => {
    // Normal window - open tab in it
    if (!current.incognito) {
      if (!nearActive) {
        browser.tabs.create({ url }, aCallback);
      } else {
        browser.tabs.query({ active: true, currentWindow: true }, (tabs) => {
          browser.tabs.create({ url, index: (tabs[0] ? tabs[0].index + 1 : undefined) }, aCallback);
        });
      }
    } else {
      // Get all windows
      browser.windows.getAll((window) => {
        let windowId = null;
        for (let i = 0; i < window.length; i += 1) {
          // We have found a normal (non-incognito) window
          if (!window[i].incognito && window[i].type === 'normal') {
            // If more normal windows were found,
            // overwrite windowId, so we get the most recent
            // opened normal window
            windowId = window[i].id;
          }
        }
        // Create a new tab in found normal window
        if (windowId) {
          if (!nearActive) {
            browser.tabs.create({ windowId, url, active: true }, aCallback);
          } else {
            browser.tabs.query({ active: true, windowId }, (tabs) => {
              browser.tabs.create({
                windowId,
                url,
                index: (tabs[0] ? tabs[0].index + 1 : undefined),
                active: true,
              }, aCallback);
              browser.windows.update(windowId, { focused: true });
            });
          }
          browser.windows.update(windowId, { focused: true });
        } else {
          // Normal window is not currently opened,
          // so create a new one
          browser.tabs.create({ url }, aCallback);
        }
      });
    }
  });
}

// Reload already opened tab
// Input:
//   tabId: integer - id of the tab which should be reloaded
function reloadTab(tabId) {
  const listener = function (aTabId, changeInfo, tab) {
    if (changeInfo.status === 'complete' &&
          tab.status === 'complete') {
      setTimeout(() => {
        browser.runtime.sendMessage({ command: 'reloadcomplete' });
        browser.tabs.onUpdated.removeListener(listener);
      }, 2000);
    }
  };

  let tabIdNum = tabId;
  if (typeof tabIdNum === 'string') {
    tabIdNum = parseInt(tabId, 10);
  }

  browser.tabs.reload(tabIdNum, { bypassCache: true }, () => {
    browser.tabs.onUpdated.addListener(listener);
  });
}

// Implement blocking via the Chrome webRequest API.
// Stores url, whitelisting, and blocking info for a tabid+frameid
const frameData = {
  // Returns the data object for the frame with ID frameId on the tab with
  // ID tabId. If frameId is not specified, it'll return the data for all
  // frames on the tab with ID tabId. Returns undefined if tabId and frameId
  // are not being tracked.
  get(tabId, frameId) {
    if (frameId !== undefined) {
      return (frameData[tabId] || {})[frameId];
    }
    return frameData[tabId];
  },

  // Record that |tabId|, |frameId| points to |url|.
  record(tabId, frameId, url) {
    const fd = frameData;
    if (!fd[tabId]) {
      fd[tabId] = {};
    }
    fd[tabId][frameId] = {
      url,
      // Cache these as they'll be needed once per request
      domain: parseUri(url).hostname,
      resources: {},
    };
    if (frameId > 0) {
      // process a sub document / iframe, slightly differently
      fd[tabId][frameId].whitelisted = pageIsWhitelisted(
        url,
        ElementTypes.subdocument, fd[tabId][0].domain,
      );
    } else {
      fd[tabId][frameId].whitelisted = pageIsWhitelisted(url);
    }
  },

  // Watch for requests for new tabs and frames, and track their URLs.
  // Inputs: details: object from onBeforeRequest callback
  // Returns false if this request's tab+frame are not trackable.
  track(details) {
    const fd = frameData;
    const { tabId } = details;

    // A hosted app's background page
    if (tabId === -1) {
      return false;
    }

    if (details.type === 'main_frame') { // New tab
      delete fd[tabId];
      fd.record(tabId, 0, details.url);
      fd[tabId].blockCount = 0;
      if (loggingEnable) {
        log('\n-------', fd.get(tabId, 0).domain, ': loaded in tab', tabId, '--------\n\n');
      }

      return true;
    }

    // Request from a tab opened before AdBlock started, or from a
    // chrome:// tab containing an http:// iframe
    if (!fd[tabId]) {
      if (loggingEnable) {
        log('[DEBUG]', 'Ignoring unknown tab:', tabId, details.frameId, details.url);
      }

      return false;
    }

    // Some times e.g. Youtube create empty iframes via JavaScript and
    // inject code into them.  So requests appear from unknown frames.
    // Treat these frames as having the same URL as the tab.
    const potentialEmptyFrameId = (details.type === 'sub_frame' ? details.parentFrameId : details.frameId);
    if (undefined === fd.get(tabId, potentialEmptyFrameId)) {
      fd.record(tabId, potentialEmptyFrameId, fd.get(tabId, 0).url);
      if (loggingEnable) {
        log('[DEBUG]', 'Null frame', tabId, potentialEmptyFrameId, "found; giving it the tab's URL.");
      }
    }

    if (details.type === 'sub_frame') { // New frame
      fd.record(tabId, details.frameId, details.url);
      if (loggingEnable) {
        log('[DEBUG]', '=========== Tracking frame', tabId, details.parentFrameId, details.frameId, details.url);
      }
    }

    return true;
  },

  // Save a resource for the resource blocker.
  storeResource(tabId, frameId, url, elType, frameDomain) {
    if (!getSettings().show_advanced_options) {
      return;
    }
    const data = frameData.get(tabId, frameId);
    if (data !== undefined &&
            data.resources) {
      data.resources[`${elType}:|:${url}:|:${frameDomain}`] = null;
    }
  },

  removeTabId(tabId) {
    delete frameData[tabId];
  },
};

// When a request starts, perhaps block it.
function onBeforeRequestHandler(details) {
  if (adblockIsPaused()) {
    return { cancel: false };
  }

  const newDetails = details;
  newDetails.url = getUnicodeUrl(newDetails.url);

  if (!frameData.track(newDetails)) {
    return { cancel: false };
  }

  const { tabId, type: reqType } = newDetails;

  const topFrame = frameData.get(tabId, 0);
  const subFrame = (newDetails.frameId !== 0 ? frameData.get(tabId, newDetails.frameId) : null);

  // If top frame is whitelisted, don't process anything
  if (topFrame.whitelisted) {
    if (loggingEnable) {
      log('[DEBUG]', 'Ignoring whitelisted tab', tabId, newDetails.url.substring(0, 100));
    }

    return { cancel: false };
    // If request comes from whitelisted subFrame and
  // top frame is not whitelisted, don't process the request
  } else if (subFrame && subFrame.whitelisted) {
    if (loggingEnable) {
      log('[DEBUG]', 'Ignoring whitelisted frame', tabId, newDetails.url.substring(0, 100));
    }

    return { cancel: false };
  }

  // For most requests, Chrome and we agree on who sent the request: the frame.
  // But for iframe loads, we consider the request to be sent by the outer
  // frame, while Chrome claims it's sent by the new iframe.  Adjust accordingly.
  const requestingFrameId = (reqType === 'sub_frame' ? newDetails.parentFrameId : newDetails.frameId);

  const elType = ElementTypes.fromOnBeforeRequestType(reqType);

  // May the URL be loaded by the requesting frame?
  const frameDomain = frameData.get(tabId, requestingFrameId).domain;

  // If |matchGeneric| is null, don't test request against blocking generic rules
  const matchGeneric = myfiltersObj.blocking.whitelist.matches(
    topFrame.url,
    ElementTypes.genericblock,
    topFrame.url,
  );
  let blocked;
  if (getSettings().data_collection) {
    const blockedData = myfiltersObj.blocking.matches(
      newDetails.url,
      elType,
      frameDomain,
      true,
      true,
      matchGeneric,
    );
    if (blockedData !== false) {
      DataCollection.addItem(blockedData.text);
      ({ blocked } = blockedData.blocked);
    } else {
      blocked = blockedData;
    }
  } else {
    blocked = myfiltersObj.blocking.matches(
      newDetails.url,
      elType,
      frameDomain,
      false,
      false,
      matchGeneric,
    );
  }

  frameData.storeResource(tabId, requestingFrameId, newDetails.url, elType, frameDomain);

  if (blocked && frameDomain === 'www.hulu.com') {
    if (frameData.get(tabId, 0).domain !== 'www.hulu.com'
        && /ads\.hulu\.com/.test(newDetails.url)) {
      // good enough
      blocked = false;
    }
  }

  /* eslint-disable-next-line no-bitwise */
  const canPurge = (elType & (ElementTypes.image | ElementTypes.subdocument | ElementTypes.object));
  if (canPurge && blocked) {
    // frameUrl is used by the recipient to determine whether they're the frame who should
    // receive this or not.  Because the #anchor of a page can change without navigating
    // the frame, ignore the anchor when matching.
    const frameUrl = frameData.get(tabId, requestingFrameId).url.replace(/#.*$/, '');
    const data = {
      command: 'purge-elements', tabId, frameUrl, url: newDetails.url, elType,
    };
    browser.tabs.sendMessage(tabId, data);
  }

  if (blocked) {
    blockCounts.recordOneAdBlocked(tabId);
    updateBadge(tabId);
  }

  if (loggingEnable) {
    log('[DEBUG]', 'Block result', blocked, reqType, frameDomain, newDetails.url.substring(0, 100));
  }

  if (blocked && elType === ElementTypes.subdocument) {
    return { redirectUrl: 'about:blank' };
  }

  return { cancel: blocked };
}

const handleEarlyOpenedTabs = function (tabs) {
  if (!tabs) {
    return;
  }

  log('Found', tabs.length, 'tabs that were already opened');
  for (let i = 0; i < tabs.length; i += 1) {
    const currentTab = tabs[i];
    const tabId = currentTab.id;
    if (!frameData.get(tabId)) { // unknown tab
      currentTab.url = getUnicodeUrl(currentTab.url);
      frameData.track({ url: currentTab.url, tabId, type: 'main_frame' });
    }

    updateBadge(tabId);
    // TODO: once we're able to get the parentFrameId, call
    // browser.webNavigation.getAllFrames to 'load' the subframes
  }
};

const handlerBehaviorChanged = function () {
  try {
    browser.webRequest.handlerBehaviorChanged();
  } catch (ex) {
    log(ex);
  }
};

// Popup blocking
function onCreatedNavigationTargetHandler(details) {
  if (adblockIsPaused()) {
    return;
  }
  const opener = frameData.get(details.sourceTabId, details.sourceFrameId);
  if (opener === undefined) {
    return;
  }
  if (frameData.get(details.sourceTabId, 0).whitelisted) {
    return;
  }
  // Change to opener's url in so that it would still be tested against the
  // blocking filter's regex rule.
  let detailsUrl = details.url;
  if (detailsUrl === 'about:blank') {
    detailsUrl = opener.url;
  }
  const url = getUnicodeUrl(detailsUrl);
  const match = myfiltersObj.blocking.matches(url, ElementTypes.popup, opener.domain);
  if (match) {
    browser.tabs.remove(details.tabId);
    blockCounts.recordOneAdBlocked(details.sourceTabId);
    updateBadge(details.sourceTabId);
  }
  frameData.storeResource(
    details.sourceTabId,
    details.sourceFrameId,
    url,
    ElementTypes.popup,
    opener.domain,
  );
}

// CUSTOM FILTERS

// Get the custom filters text as a \n-separated text string.
const getCustomFiltersText = function (callback) {
  if (typeof callback !== 'function') {
    return;
  }

  browser.storage.local.get('custom_filters', (response) => {
    callback(response.custom_filters);
  });
};

// Set the custom filters to the given \n-separated text string, and
// rebuild the filterset.
// Inputs: filters:string the new filters.
const setCustomFiltersText = function (filters) {
  browser.storage.local.set({ custom_filters: filters });
  browser.runtime.sendMessage({ command: 'filters_updated' });
  myfiltersObj.rebuild();
};

// Get the user enterred exclude filters text as a \n-separated text string.
const getExcludeFiltersText = function (callback) {
  if (typeof callback !== 'function') {
    return;
  }

  browser.storage.local.get('exclude_filters', (response) => {
    callback(response.exclude_filters);
  });
};
// Set the exclude filters to the given \n-separated text string, and
// rebuild the filterset.
// Inputs: filters:string the new filters.
const setExcludeFilters = function (filters) {
  let newFilters = filters.trim();
  newFilters = newFilters.replace(/\n\n/g, '\n');
  browser.storage.local.set({ exclude_filters: newFilters });
  FilterNormalizer.setExcludeFilters(newFilters);
  updateSubscriptionsNow();
};
// Add / concatenate the exclude filter to the existing excluded filters, and
// rebuild the filterset.
// Inputs: filter:string the new filter.
const addExcludeFilter = function (filter) {
  getExcludeFiltersText((currentExcludedFilters) => {
    if (currentExcludedFilters) {
      setExcludeFilters(`${currentExcludedFilters}\n${filter}`);
    } else {
      setExcludeFilters(filter);
    }
  });
};

// Removes a custom filter entry.
// Inputs: host:domain of the custom filters to be reset.
const removeCustomFilter = function (host) {
  getCustomFiltersText((text) => {
    const customFiltersArr = text ? text.split('\n') : [];
    const newCustomFiltersArr = [];
    const identifier = host;

    for (let i = 0; i < customFiltersArr.length; i += 1) {
      const entry = customFiltersArr[i];
      // Make sure that the identifier is at the start of the entry
      if (entry.indexOf(identifier) !== 0) {
        newCustomFiltersArr.push(entry);
      }
    }

    const newText = newCustomFiltersArr.join('\n');
    setCustomFiltersText(newText.trim());
  });
};

let countCache = {};
// countCache singleton.
const countCacheInit = function () {
  browser.storage.local.get('custom_filter_count', (response) => {
    countCache = (function countCacheFunction(countMap) {
      let cache = countMap;
      // Update custom filter count stored in localStorage
      const updateCustomFilterCount = function () {
        const dataToStore = {};
        dataToStore.custom_filter_count = cache;
        browser.storage.local.set(dataToStore);
      };

      return {
        // Update custom filter count cache and value stored in localStorage.
        // Inputs: newCountMap:count map - count map to replace existing count cache
        updateCustomFilterCountMap(newCountMap) {
          let newerCountMap = newCountMap;
          if (newCountMap) {
            newerCountMap = JSON.parse(JSON.stringify(newCountMap));
          }
          cache = newerCountMap || cache;
          updateCustomFilterCount();
        },
        // Remove custom filter count for host
        // Inputs: host:string - url of the host
        removeCustomFilterCount(host) {
          if (host && cache[host]) {
            delete cache[host];
            updateCustomFilterCount();
          }
        },
        // Get current custom filter count for a particular domain
        // Inputs: host:string - url of the host
        getCustomFilterCount(host) {
          return cache[host] || 0;
        },
        // Add 1 to custom filter count for the filters domain.
        // Inputs: filter:string - line of text to be added to custom filters.
        addCustomFilterCount(filter) {
          const host = filter.split('##')[0];
          cache[host] = this.getCustomFilterCount(host) + 1;
          updateCustomFilterCount();
        },
      };
    }(response.custom_filter_count || {}));
  });
};

// Entry point for customize.js, used to update custom filter count cache.
const updateCustomFilterCountMap = function (newCountMap) {
  countCache.updateCustomFilterCountMap(newCountMap);
};

const removeCustomFilterForHost = function (host) {
  if (countCache.getCustomFilterCount(host)) {
    removeCustomFilter(host);
    countCache.removeCustomFilterCount(host);
  }
};

// Currently, Firefox doesn't allow the background page to use alert() or confirm(),
// so some JavaScript is injected into the active tab, which does the confirmation for us.
// If the user confirms the removal of the entries, then they are removed, and the page reloaded.
const confirmRemovalOfCustomFiltersOnHost = function (host, activeTab) {
  const customFilterCount = countCache.getCustomFilterCount(host);
  const confirmationText = translate('confirm_undo_custom_filters', [customFilterCount, host]);
  const messageListenerFN = function (request) {
    browser.runtime.onMessage.removeListener(messageListenerFN);
    if (request === `remove_custom_filters_on_host${host}:true`) {
      removeCustomFilterForHost(host);
      browser.tabs.executeScript(activeTab.id, { code: 'location.reload();' });
    }
  };

  browser.runtime.onMessage.addListener(messageListenerFN);
  const codeToExecute = `var host = "${host}"; var confirmResponse = confirm("${confirmationText}"); browser.runtime.sendMessage("remove_custom_filters_on_host" + host + ":" + confirmResponse); `;
  const details = { allFrames: false, code: codeToExecute };
  browser.tabs.executeScript(activeTab.id, details);
};

// UNWHITELISTING

// Look for a custom filter that would whitelist options.url,
// and if any exist, remove the first one.
// Inputs: url:string - a URL that may be whitelisted by a custom filter
// Returns: true if a filter was found and removed; false otherwise.
const tryToUnwhitelist = function (url, callback) {
  let newUrl = url.replace(/#.*$/, ''); // Whitelist ignores anchors
  getCustomFiltersText((customFiltersText) => {
    const customFilters = customFiltersText.split('\n');
    for (let i = 0; i < customFilters.length; i += 1) {
      let text = customFilters[i];
      const whitelist = text.search(/@@\*\$document,domain=~/);
      // Blacklist site, which is whitelisted by global @@*&document,domain=~ filter
      if (whitelist > -1) {
        // Remove protocols
        [newUrl] = newUrl.replace(/((http|https):\/\/)?(www.)?/, '').split(/[/?#]/);
        text = `${text}|~${newUrl}`;
        customFilters.splice(i, 1); // Remove the old filter text
        customFilters.push(text); // add the new filter text to original array
        const newText = customFilters.join('\n');
        setCustomFiltersText(newText);
        if (typeof callback === 'function') {
          callback(true);
        }
        browser.runtime.sendMessage({
          message: 'try_to_unwhitelist_response',
          returnCode: true,
        });
        return true;
      }
      if (Filter.isWhitelistFilter(text)) {
        let filter;
        try {
          filter = PatternFilter.fromText(text);
        } catch (ex) {
          filter = false;
        }

        if (filter && filter.matches(newUrl, ElementTypes.document, false)) {
          customFilters.splice(i, 1); // Remove this whitelist filter text
          const newText = customFilters.join('\n');
          setCustomFiltersText(newText);
          if (typeof callback === 'function') {
            callback(true);
          }
          browser.runtime.sendMessage({
            message: 'try_to_unwhitelist_response',
            returnCode: true,
          });
          return true;
        }
      }
    }

    if (typeof callback === 'function') {
      callback(false);
    }
    browser.runtime.sendMessage({
      message: 'try_to_unwhitelist_response',
      returnCode: false,
    });
    return false;
  });
};

// Get or set if AdBlock is paused
// Inputs: newValue (optional boolean): if true, AdBlock will be paused, if
//                  false, AdBlock will not be paused.
// Returns: undefined if newValue was specified, otherwise it returns true
//          if paused, false otherwise.
let adblockIsPausedBool = false;
const adblockIsPaused = function (newValue) {
  if (newValue === undefined) {
    return (adblockIsPausedBool === true);
  }

  adblockIsPausedBool = newValue;
  return undefined;
};

// Get if AdBlock is paused
// called from content scripts
// Returns: true if paused, false otherwise.
const isAdblockPaused = function () {
  return adblockIsPaused();
};

// INFO ABOUT CURRENT PAGE

// Get interesting information about the current tab.
// Inputs:
//   callback: function(info).
//   info object passed to callback: {
//     tab: Tab object
//     whitelisted: bool - whether the current tab's URL is whitelisted.
//     disabled_site: bool - true if the url is e.g. about:blank or the
//                           Extension Gallery, where extensions don't run.
//     total_blocked: int - # of ads blocked since install
//     tab_blocked: int - # of ads blocked on this tab
//     display_stats: bool - whether block counts are displayed on button
//     display_menu_stats: bool - whether block counts are displayed on the popup menu
//   }
// Returns: null (asynchronous)
const getCurrentTabInfo = function (callback) {
  const processTab = function (tab) {
    if (!tab) {
      return;
    }
    const result = {
      tab,
      display_stats: getSettings().display_stats,
      display_menu_stats: getSettings().display_menu_stats,
    };

    if (!tab.url) {
      result.disabled_site = true;
      result.tab_blocked = 0;
    } else {
      result.tab.unicodeUrl = getUnicodeUrl(tab.url);
      result.disabled_site = pageIsUnblockable(tab.unicodeUrl);
      if (!result.disabled_site) {
        result.whitelisted = pageIsWhitelisted(tab.unicodeUrl);
      }
    }

    if (getSettings().youtube_channel_whitelist &&
        parseUri(tab.unicodeUrl).hostname === 'www.youtube.com') {
      result.youTubeChannelName = frameData.get(tab.id, 0).youTubeChannelName;
    }

    result.total_blocked = blockCounts.getTotalAdsBlocked();
    if (tab.url && tab.id) {
      result.tab_blocked = blockCounts.getTotalAdsBlocked(tab.id);
      if (typeof callback === 'function') {
        callback(result);
      }
    } else if (typeof callback === 'function') {
      callback(result);
    }
  };

  if (currentTabId) {
    if (typeof currentTabId === 'string') {
      currentTabId = Number.parseInt(currentTabId, 10);
    }
    browser.tabs.get(currentTabId, processTab);
  } else {
    browser.tabs.query({ active: true, currentWindow: true }, (tabs) => {
      if (!tabs) {
        return;
      }

      if (tabs.length === 0) {
        return;
      } // For example: only the background devtools or a popup are opened

      const tab = tabs[0];
      processTab(tab);
    });
  }
};
let currentTabId;
browser.runtime.onMessage.addListener((request, sender, sendResponse) => {
  if (!(request.message === 'getCurrentTabInfo')) {
    return;
  }
  if (request.tabId) {
    currentTabId = request.tabId;
  } else {
    currentTabId = undefined;
  }
  getCurrentTabInfo((info) => {
    if (info) {
      const updatedInfo = info;
      updatedInfo.settings = getSettings();
      updatedInfo.paused = adblockIsPaused();
      const eligibleForUndo = !updatedInfo.paused
        && (updatedInfo.disabled_site || !updatedInfo.whitelisted);
      updatedInfo.customFilterCount = 0;
      if (eligibleForUndo && updatedInfo.tab) {
        const { host } = parseUri(updatedInfo.tab.unicodeUrl);
        const urlToCheckForUndo = updatedInfo.disabled_site ? undefined : host;
        updatedInfo.customFilterCount = countCache.getCustomFilterCount(urlToCheckForUndo);
      }
      browser.runtime.sendMessage({
        message: 'getCurrentTabInfoResponse',
        info: updatedInfo,
      });
    }
  });
  sendResponse({});
});

// Returns true if anything in whitelist matches the_domain.
//   url: the url of the page
//   type: one out of ElementTypes, default ElementTypes.document,
//         to check what the page is whitelisted for: hiding rules or everything
const pageIsWhitelisted = function (url, type, frameDomain) {
  if (!url) {
    return true;
  }

  let cleanedUrl = getUnicodeUrl(url);
  cleanedUrl = cleanedUrl.replace(/#.*$/, ''); // Remove anchors
  let newType = type;
  if (!type) {
    newType = ElementTypes.document;
  }
  if (!myfiltersObj || !myfiltersObj.blocking || !myfiltersObj.blocking.whitelist) {
    return false;
  }
  const { whitelist } = myfiltersObj.blocking;
  if (frameDomain) {
    return whitelist.matches(cleanedUrl, newType, frameDomain, false);
  }
  return whitelist.matches(cleanedUrl, newType, parseUri(cleanedUrl).hostname, false);
};

const setBrowserActions = function (options) {
  if (typeof browser.browserAction.setIcon === 'undefined') {
    return;
  }

  const iconCallback = function () {
    if (browser.runtime.lastError) {
      return;
    }
    if (typeof browser.browserAction.setBadgeText === 'undefined' ||
            typeof browser.browserAction.setBadgeBackgroundColor === 'undefined') {
      return;
    }
    browser.browserAction.setBadgeText({ text: options.badgeText, tabId: options.tabId });
    browser.browserAction.setBadgeBackgroundColor({ color: options.color });
  };
  browser.browserAction.setIcon({ tabId: options.tabId, path: options.iconPaths }, iconCallback);
};

const updateBadge = function (tabId) {
  const display = getSettings().display_stats;
  const mainFrame = frameData.get(tabId, 0);
  // mainFrame is undefined if the tab is a new one, so no use updating badge.
  if (!mainFrame) {
    return;
  }

  const isBlockable = !pageIsUnblockable(mainFrame.url) && !pageIsWhitelisted(mainFrame.url) && !/chrome\/newtab/.test(mainFrame.url);

  if ((mainFrame && isBlockable) && !adblockIsPaused()) {
    const browsersBadgeOptions = {};
    browsersBadgeOptions.tabId = tabId;
    browsersBadgeOptions.color = '#555';
    const adsBlocked = blockCounts.getTotalAdsBlocked(tabId);
    let badgeText = adsBlocked.toString();
    if (!display || badgeText === '0') {
      badgeText = ''; // Only show the user when we've done something useful
    }

    browsersBadgeOptions.badgeText = badgeText;
    browsersBadgeOptions.iconPaths = { 19: 'img/icon20.png', 38: 'img/icon40.png' };
    setBrowserActions(browsersBadgeOptions);
  }
};

// Set the button image and context menus according to the URL
// of the current tab.
const updateButtonUIAndContextMenus = function () {
  function setContextMenus(info) {
    if (Fennec) {
      return;
    }
    browser.contextMenus.removeAll();
    if (!getSettings().show_context_menu_items) {
      return;
    }

    if (adblockIsPaused() || info.whitelisted || info.disabled_site) {
      return;
    }

    function addMenu(title, callback) {
      browser.contextMenus.create({
        id: title,
        title,
        contexts: ['all'],
        onclick(clickdata, tab) {
          callback(tab, clickdata);
        },
      });
    }

    addMenu(translate('block_this_ad'), (tab, clickdata) => {
      emitPageBroadcast(
        { fn: 'topOpenBlacklistUI', options: { info: clickdata } },
        { tab },
      );
    });

    addMenu(translate('block_an_ad_on_this_page'), (tab) => {
      emitPageBroadcast(
        { fn: 'topOpenBlacklistUI', options: { nothing_clicked: true } },
        { tab },
      );
    });

    const host = getUnicodeDomain(parseUri(info.tab.unicodeUrl).host);
    const customFilterCount = countCache.getCustomFilterCount(host);
    if (customFilterCount) {
      addMenu(translate('undo_last_block'), (tab) => {
        confirmRemovalOfCustomFiltersOnHost(host, tab);
      });
    }
  }

  function setBrowserButton(info) {
    const browsersBadgeOptions = {};
    browsersBadgeOptions.tabId = info.tab.id;
    browsersBadgeOptions.color = '#555';
    browsersBadgeOptions.badgeText = '';
    if (adblockIsPaused()) {
      browsersBadgeOptions.iconPaths = { 19: 'img/icon20-grayscale.png', 38: 'img/icon40-grayscale.png' };
      setBrowserActions(browsersBadgeOptions);
    } else if (info.disabled_site &&
        !/^chrome-extension:.*pages\/install\//.test(info.tab.unicodeUrl)) {
      // Show non-disabled icon on the installation-success page so it
      // users see how it will normally look. All other disabled pages
      // will have the gray one
      browsersBadgeOptions.iconPaths = { 19: 'img/icon20-grayscale.png', 38: 'img/icon40-grayscale.png' };
      setBrowserActions(browsersBadgeOptions);
    } else if (info.whitelisted) {
      browsersBadgeOptions.iconPaths = { 19: 'img/icon20-whitelisted.png', 38: 'img/icon40-whitelisted.png' };
      setBrowserActions(browsersBadgeOptions);
    } else {
      updateBadge(info.tab.id);
    }
  }

  getCurrentTabInfo((info) => {
    setContextMenus(info);
    setBrowserButton(info);
  });
};

// Returns true if the url cannot be blocked
const pageIsUnblockable = function (url) {
  if (!url) {
    return true;
  }
  const { protocol: scheme, pathname } = parseUri(url);
  return ((scheme !== 'http:' && scheme !== 'https:' && scheme !== 'feed:') || (pathname && pathname.includes('.pdf')));
};

// MYFILTERS PASSTHROUGHS

// Rebuild the filterset based on the current settings and subscriptions.
const updateFilters = function () {
  myfiltersObj.rebuild();
};

// Fetch the latest version of all subscribed lists now.
const updateSubscriptionsNow = function () {
  myfiltersObj.checkFilterUpdates(true);
};

// Returns map from id to subscription object.  See filters.js for
// description of subscription object.
const getSubscriptionsMinusText = function () {
  const result = {};
  for (const id in myfiltersObj.subscriptions) {
    result[id] = {};
    for (const attr in myfiltersObj.subscriptions[id]) {
      if (attr !== 'text') {
        result[id][attr] = myfiltersObj.subscriptions[id][attr];
      }
    }
  }

  return result;
};

// Returns map from id to subscription object.  See filters.js for
// description of subscription object.
// returns all subscription data which will be a large object
const getSubscriptions = function () {
  const result = {};
  for (const id in myfiltersObj.subscriptions) {
    result[id] = {};
    for (const attr in myfiltersObj.subscriptions[id]) {
      result[id][attr] = myfiltersObj.subscriptions[id][attr];
    }
  }

  return result;
};

// Subscribes to a filter subscription.
// Inputs: id: id to which to subscribe.  Either a well-known
//             id, or "url:xyz" pointing to a user-specified list.
//         requires: the id of a list if it is a supplementary list,
//                   or null if nothing required
// Returns: null, upon completion
const subscribe = function (options) {
  myfiltersObj.changeSubscription(options.id, {
    subscribed: true,
    requiresList: options.requires,
    title: options.title,
  });
};

// Unsubscribes from a filter subscription.
// Inputs: id: id from which to unsubscribe.
//         del: (bool) if the filter should be removed or not
// Returns: null, upon completion.
const unsubscribe = function (options) {
  myfiltersObj.changeSubscription(options.id, {
    subscribed: false,
    deleteMe: (options.del ? true : undefined),
  });
};

// DEBUG INFO
// Create the debug info for the textbox or the bug report
const getDebugInfo = function (callback) {
  // An object, which contains info about AdBlock like
  // subscribed filter lists, settings and other settings
  const theDebugInfo = {
    filter_lists: [],
    settings: [],
    other_info: [],
    custom_filters: '',
  };
  // Get subscribed filter lists
  const subs = getSubscriptionsMinusText();
  theDebugInfo.filter_lists.push('=== Filter Lists ===');
  theDebugInfo.filter_lists.push('');
  for (const id in subs) {
    if (subs[id].subscribed) {
      theDebugInfo.filter_lists.push(id);
      theDebugInfo.filter_lists.push(`  last updated: ${new Date(subs[id].lastUpdate).toUTCString()}`);
      theDebugInfo.filter_lists.push('');
    }
  }
  // Format info about filter lists properly
  theDebugInfo.filter_lists = theDebugInfo.filter_lists.join('\n');
  // Get settings
  theDebugInfo.settings.push('=== Settings ===');
  const settings = getSettings();
  for (const setting in settings) {
    theDebugInfo.settings.push(`${setting}: ${getSettings()[setting]}`);
  }

  theDebugInfo.settings = theDebugInfo.settings.join('\n');

  theDebugInfo.other_info.push('');
  theDebugInfo.other_info.push(`AdBlock version number: ${browser.runtime.getManifest().version}`);
  theDebugInfo.other_info.push(`browser =${STATS.browser}`);
  theDebugInfo.other_info.push(`browserVersion = ${STATS.browserVersion}`);
  theDebugInfo.other_info.push(`osVersion =${STATS.osVersion}`);
  theDebugInfo.other_info.push(`os =${STATS.os}`);
  theDebugInfo.other_info.push(`user language =${determineUserLanguage()}`);
  theDebugInfo.other_info.push(`Fennec (Mobile) = ${Fennec}`);
  theDebugInfo.other_info.push('');

  if (localStorage &&
      localStorage.length) {
    theDebugInfo.other_info.push('local storage data');
    theDebugInfo.other_info.push(`length =${localStorage.length}`);
    let inx = 1;
    for (const key in localStorage) {
      theDebugInfo.other_info.push(`key ${inx} = ${key}`);
      inx += 1;
    }
  } else {
    theDebugInfo.other_info.push('no local storage data');
  }

  theDebugInfo.other_info.push('');
  // Get total pings
  browser.storage.local.get('total_pings', (storageResponse) => {
    theDebugInfo.other_info.push(`total_pings =${storageResponse.total_pings || 0}`);

    // Now, add exclude filters (if there are any)
    const excludeFiltersKey = 'exclude_filters';
    browser.storage.local.get(excludeFiltersKey, (secondResponse) => {
      if (secondResponse && secondResponse[excludeFiltersKey]) {
        theDebugInfo.other_info.push('==== Exclude Filters ====');
        theDebugInfo.other_info.push(secondResponse[excludeFiltersKey]);
        theDebugInfo.other_info.push('');
      }
      // Now, add the migration messages (if there are any)
      const migrateLogMessageKey = 'migrateLogMessageKey';
      browser.storage.local.get(migrateLogMessageKey, (thirdResponse) => {
        if (thirdResponse && thirdResponse[migrateLogMessageKey]) {
          theDebugInfo.other_info.push('');
          const messages = thirdResponse[migrateLogMessageKey].split('\n');
          for (let i = 0; i < messages.length; i += 1) {
            const key = `migration_message_${i}`;
            theDebugInfo.other_info.push(`${key} : ${messages[i]}`);
          }

          theDebugInfo.other_info.push('');
        }

        const malwareKey = 'malware-notification';
        browser.storage.local.get(migrateLogMessageKey, (fourthResponse) => {
          if (fourthResponse && fourthResponse[malwareKey]) {
            // We need to hardcode malware-notification setting,
            // because it isn't included in settingsObj object, but just in localStorage
            theDebugInfo.other_info.push(`malware-notification: ${fourthResponse[malwareKey]}`);
          }

          const customFiltersKey = 'custom_filters';
          browser.storage.local.get(customFiltersKey, (fifthResponse) => {
            if (fifthResponse && fifthResponse[customFiltersKey]) {
              theDebugInfo.custom_filters = fifthResponse[customFiltersKey];
            }

            theDebugInfo.other_info = theDebugInfo.other_info.join('\n');
            if (typeof callback === 'function') {
              callback(theDebugInfo);
            }
          });
        });
      });
    });
  });
};

// Inputs: options object containing:
//           domain:string the domain of the calling frame.
browser.runtime.onMessage.addListener((request, sender, sendResponse) => {
  if (request.message !== 'debug_report_elemhide') {
    return;
  }

  if (!window.frameData) {
    return;
  }

  if (!sender) {
    return;
  }

  const { selector, matches } = request.opts;
  const frameDomain = parseUri(sender.url || sender.tab.url).hostname;
  frameData.storeResource(sender.tab.id, sender.frameId || 0, selector, 'selector', frameDomain);

  const data = frameData.get(sender.tab.id, sender.frameId || 0);
  if (data) {
    if (loggingEnable) {
      log(data.domain, ': hiding rule', selector, 'matched:\n', matches);
    }

    blockCounts.recordOneAdBlocked(sender.tab.id);
    updateBadge(sender.tab.id);
  }

  sendResponse({});
});

// BGcall DISPATCH
(function bgCallDispatch() {
  browser.runtime.onMessage.addListener((request, sender, sendResponse) => {
    if (request.command !== 'call') {
      return;
    } // not for us
    const fn = window[request.fn];
    if (typeof fn !== 'function') {
      log('fn', fn, 'not found on background page, request: ', request);
      return;
    }
    request.args.push(sender);
    const result = fn.apply(window, request.args, sender);
    sendResponse(result);
  });
}());

// These functions are usually only called by content scripts.

// Add a new custom filter entry.
// Inputs: filter:string line of text to add to custom filters.
// Invokes callback (if provided) with null if succesfull, otherwise an exception
// Since this function is async, if calling from a content script the validateFilter()
// function should be called first.
const addCustomFilter = function (filter, callback) {
  const newFilter = filter;
  getCustomFiltersText((text) => {
    try {
      if (FilterNormalizer.normalizeLine(newFilter)) {
        if (Filter.isSelectorFilter(newFilter)) {
          countCache.addCustomFilterCount(newFilter);
          updateButtonUIAndContextMenus();
        }
        let newText = text;
        if (!newText) {
          newText = '';
        } else {
          newText += '\n';
        }
        const customFilters = newText + newFilter;
        setCustomFiltersText(customFilters);
        if (typeof callback === 'function') {
          callback(null);
        }
      }

      if (typeof callback === 'function') {
        callback('This filter is unsupported');
      }
    } catch (ex) {
      // convert to a string so that it can be passed
      // back to content scripts
      if (typeof callback === 'function') {
        callback(ex.toString());
      }
    }
  });
};

const validateFilter = function (filter, returnException) {
  const response = FilterNormalizer.validateLine(filter, returnException);
  if (returnException && response && response.exception) {
    return JSON.stringify(response);
  }

  return response;
};

// Creates a custom filter entry that whitelists a given page
// Inputs: url:string url of the page
// Returns: null if successful, otherwise an exception
const createPageWhitelistFilter = function (url) {
  const newUrl = url.replace(/#.*$/, ''); // Remove anchors
  const parts = newUrl.match(/^([^?]+)(\??)/); // Detect querystring
  const hasQuerystring = parts[2];
  const filter = `@@|${parts[1]}${hasQuerystring ? '?' : '|'}$document`;
  return addCustomFilter(filter);
};

// Creates a custom filter entry that whitelists a YouTube channel
// Inputs: url:string url of the page
// Returns: null if successful, otherwise an exception
const createWhitelistFilterForYoutubeChannel = function (url) {
  let ytChannel;
  if (/ab_channel=/.test(url)) {
    [, ytChannel] = url.match(/ab_channel=([^]*)/);
  } else {
    ytChannel = url.split('/').pop();
  }

  if (ytChannel) {
    const filter = `@@|https://www.youtube.com/*${ytChannel}|$document`;
    return addCustomFilter(filter);
  }
  return undefined;
};

// Listen for the message from the ytchannel.js content script
browser.runtime.onMessage.addListener((message, sender, sendResponse) => {
  if (message.command === 'updateYouTubeWhitelistFilters') {
    const response = updateYouTubeWhitelistFilters(message.args);
    sendResponse(response);
  } else if (message.command === 'updateYouTubeChannelName') {
    const fd = frameData.get(sender.tab.id, 0);
    fd.youTubeChannelName = message.args;
    sendResponse({});
  }
});

// Creates a custom filter entry that whitelists a YouTube channel
// Inputs: titles:string array of the YouTube Channels to Whitelist.
// The titles / channel names in the input string should be parsed, and
// URLEncoded prior to calling this function.
const updateYouTubeWhitelistFilters = function (titles) {
  if (!titles ||
      !Array.isArray(titles) ||
      titles.length < 1) {
    return true;
  }

  const newTitles = titles;
  getCustomFiltersText((text) => {
    const customFiltersArray = text ? text.split('\n') : [];
    const newCustomFiltersArray = [];
    // First, remove any old YouTube white list filters
    if (customFiltersArray) {
      for (let inx = 0; inx < customFiltersArray.length; inx += 1) {
        const filter = customFiltersArray[inx];
        if (!filter.startsWith('@@|https://www.youtube.com/*')) {
          newCustomFiltersArray.push(filter);
        }
      }
    }
    // Second, add all of the YouTube channel titles
    for (let inx = 0; inx < newTitles.length; inx += 1) {
      const title = newTitles[inx];
      if (title) {
        const filter = `@@|https://www.youtube.com/*${title}|$document`;
        newCustomFiltersArray.push(filter);
      }
    }
    // Second, add all of the YouTube channel titles
    setCustomFiltersText(newCustomFiltersArray.join('\n'));
  });
  return undefined;
};

const userStyleSheetsSupported = 'extensionTypes' in browser && 'CSSOrigin' in browser.extensionTypes;

// Inputs: options object containing:
//           domain:string the domain of the calling frame.
browser.runtime.onMessage.addListener((request, sender, sendResponse) => {
  if (!(request.message === 'get_content_script_data')) {
    return;
  }

  const options = request.opts;
  const settings = getSettings();
  const runnable = !adblockIsPaused() && !pageIsUnblockable(sender.url);
  const runningTop = runnable && !pageIsWhitelisted(sender.tab.url);
  let running = runnable && !pageIsWhitelisted(sender.url);
  let hiding = running && !pageIsWhitelisted(sender.url, ElementTypes.elemhide);

  // Don't run in frame, when top frame is whitelisted
  if (!runningTop && running) {
    running = false;
    hiding = false;
  }

  if (hiding && sender.frameId && sender.tab.id) {
    const fd = frameData.get(sender.tab.id, sender.frameId);
    if (fd && fd.whitelisted) {
      running = false;
      hiding = false;
    }
  }

  const result = {
    settings,
    runnable,
    running,
    hiding,
    Fennec,
  };

  if (hiding &&
      myfiltersObj &&
      myfiltersObj.hiding &&
      settings) {
    // If |matchGeneric| is , don't test request against hiding generic rules
    const matchGeneric = myfiltersObj.blocking.whitelist.matches(
      sender.tab.url,
      ElementTypes.generichide,
      sender.tab.url,
    );
    const selectors = myfiltersObj.hiding.filtersFor(options.domain, matchGeneric);
    if (userStyleSheetsSupported) {
      browser.tabs.insertCSS(sender.tab.id, {
        code: `${selectors.join(', ')}{display: none !important;}`,
        cssOrigin: 'user',
        frameId: sender.frameId,
        matchAboutBlank: true,
        runAt: 'document_start',
      });
      result.selectors = [];
    } else {
      result.selectors = selectors;
    }
    result.advanceSelectors = myfiltersObj.advanceHiding.advanceFiltersFor(
      options.domain,
      matchGeneric,
    );
  }

  sendResponse(result);
});

browser.runtime.onMessage.addListener((request, sender, sendResponse) => {
  if (!(request.message === 'open.options.page.subscriptions.add')) {
    return;
  }

  let opennedtabId = 0;
  const pageCompleteListener = function (tabId, info) {
    if (info.status === 'complete' && opennedtabId === tabId) {
      browser.tabs.onUpdated.removeListener(pageCompleteListener);
      browser.tabs.sendMessage(tabId, {
        message: 'addSubscription',
        title: request.title,
        url: request.url,
      });
    }
  };

  browser.tabs.onUpdated.addListener(pageCompleteListener);
  const optionsUrl = browser.extension.getURL('options/index.html?tab=1');
  browser.tabs.create({ url: optionsUrl }, (tab) => {
    opennedtabId = tab.id;
  });

  sendResponse({});
});

// Bounce messages back to content scripts.
const emitPageBroadcast = (function emitPageBroadcastFunction() {
  const injectMap = {
    topOpenWhitelistUI: {
      allFrames: false,
      include: [
        'punycode.min.js',
        'jquery/jquery-3.3.1.min.js',
        'jquery/jquery-ui.min.js',
        'uiscripts/load_css.js',
        'uiscripts/top_open_whitelist_ui.js',
      ],
    },
    topOpenBlacklistUI: {
      allFrames: false,
      include: [
        'punycode.min.js',
        'jquery/jquery-3.3.1.min.js',
        'jquery/jquery-ui.min.js',
        'uiscripts/load_css.js',
        'uiscripts/blacklisting/overlay.js',
        'uiscripts/blacklisting/clickwatcher.js',
        'uiscripts/blacklisting/elementchain.js',
        'uiscripts/blacklisting/blacklistui.js',
        'uiscripts/top_open_blacklist_ui.js',
      ],
    },
    sendContentToBack: {
      allFrames: true,
      include: [
        'uiscripts/send_content_to_back.js',
      ],
    },
  };

  // Inject the required scripts to execute fnName(parameter) in
  // the current tab.
  // Inputs: fnName:string name of function to execute on tab.
  //         fnName must exist in injectMap above.
  //         parameter:object to pass to fnName.  Must be JSON.stringify()able.
  //         injectedSoFar?:int used to recursively inject required scripts.
  const executeOnTab = function (fnName, parameter, injectedSoFar) {
    const localInjectedSoFar = injectedSoFar || 0;
    const data = injectMap[fnName];
    const details = { allFrames: data.allFrames };
    // If there's anything to inject, inject the next item and recurse.
    if (data.include.length > localInjectedSoFar) {
      details.file = data.include[localInjectedSoFar];
      browser.tabs.executeScript(undefined, details, () => {
        executeOnTab(fnName, parameter, localInjectedSoFar + 1);
      });
    } else {
      // Nothing left to inject, so execute the function.
      browser.tabs.insertCSS(undefined, { allFrames: false, file: 'jquery/css/override-page.css' });
      browser.tabs.insertCSS(undefined, { allFrames: false, file: 'jquery/css/jquery-ui.css' });
      const param = JSON.stringify(parameter);
      details.code = `${fnName}(${param});`;
      browser.tabs.executeScript(undefined, details);
    }
  };

  // The emitPageBroadcast() function
  const theFunction = function (request) {
    request.options = JSON.stringify(request.options);
    executeOnTab(request.fn, request.options);
  };

  return theFunction;
}()); // end emitPageBroadcast

// Get the current (loaded) malware domains
// Returns: an object with all of the malware domains
// will return undefined, if the user is not subscribed to the Malware 'filter list'.
const getMalwareDomains = function () {
  return myfiltersObj.getMalwareDomains();
};

// Open the resource blocker when requested from popup.
const launchResourceblocker = function (query) {
  openTab(`pages/resourceblock.html${query}`, true);
};

// Get the frameData for the 'Report an Ad' & 'Resource' page
const getFrameData = function (tabId) {
  return frameData.get(tabId);
};

// Process requests from 'Resource' page
// Determine, whether requests have been whitelisted/blocked
const processFrameData = function (fd) {
  for (const frameId in fd) {
    const frame = fd[frameId];
    const frameResources = frame.resources;
    if (frameResources) {
      for (const resource in frameResources) {
        const res = frameResources[resource];
        // We are processing selectors in resource viewer page
        if (res.elType !== 'selector') {
          res.blockedData = myfiltersObj.blocking.matches(
            res.url,
            res.elType,
            res.frameDomain,
            true,
            true,
          );
        }
      }
    }
  }

  return fd;
};

// Add previously cached requests to matchCache
// Used by 'Resource' page
const addToMatchCache = function (cache) {
  myfiltersObj.blocking.matchCache = cache;
  myfiltersObj.blocking.numCacheEntries = Object.keys(cache).length;
};

// Reset matchCache
// Used by 'Resource' page
// Returns: object with cached requests
const resetMatchCache = function () {
  const { matchCache } = myfiltersObj.blocking;
  myfiltersObj.blocking.resetMatchCache();
  return matchCache;
};

// BROWSER ACTION AND CONTEXT MENU UPDATES
if (browser.tabs) {
  browser.tabs.onUpdated.addListener((tabid, changeInfo, tab) => {
    if (tab.active && changeInfo.status) {
      updateButtonUIAndContextMenus();
    }
  });

  browser.tabs.onActivated.addListener(() => {
    updateButtonUIAndContextMenus();
  });
}

// YouTube Channel Whitelist
const runChannelWhitelist = function (tabUrl, tabId) {
  if (parseUri(tabUrl).hostname === 'www.youtube.com' &&
      getSettings().youtube_channel_whitelist &&
      !parseUri.parseSearch(tabUrl).ab_channel) {
    browser.tabs.executeScript(tabId, { file: 'ytchannel.js', runAt: 'document_start' });
  }
};

browser.tabs.onCreated.addListener((tab) => {
  if (browser.runtime.lastError) {
    return;
  }

  browser.tabs.get(tab.id, (tabs) => {
    if (browser.runtime.lastError) {
      return;
    }

    if (tabs && tabs.url && tabs.id) {
      runChannelWhitelist(tabs.url, tabs.id);
    }
  });
});

browser.tabs.onUpdated.addListener((tabId, changeInfo) => {
  if (browser.runtime.lastError) {
    return;
  }

  if (changeInfo.status) {
    browser.tabs.get(tabId, (tabs) => {
      if (browser.runtime.lastError) {
        return;
      }

      if (tabs && tabs.url && tabs.id) {
        runChannelWhitelist(tabs.url, tabs.id);
      }
    });
  }
});

browser.webNavigation.onHistoryStateUpdated.addListener((details) => {
  if (details &&
        Object.prototype.hasOwnProperty.call(details, 'frameId') &&
        Object.prototype.hasOwnProperty.call(details, 'tabId') &&
        Object.prototype.hasOwnProperty.call(details, 'url')) {
    // on some single page sites that update the URL using the History API pushState(),
    // but they don't actually load a new page, we need to get notified when this happens
    // and track these updates in the frameData object.
    const tabData = frameData.get(details.tabId, details.frameId);
    if (tabData &&
          tabData.url !== details.url) {
      const updatedDetails = details;
      updatedDetails.type = 'main_frame';
      updatedDetails.url = getUnicodeUrl(updatedDetails.url);
      frameData.track(updatedDetails);
      if (tabData.youTubeChannelName) {
        frameData.get(updatedDetails.tabId, updatedDetails.frameId).youTubeChannelName
          = tabData.youTubeChannelName;
      }
    }
  }
});


const createMalwareNotification = function () {
  if (browser &&
        browser.notifications) {
    browser.storage.local.get('malware-notification', (response) => {
      const notify = response['malware-notification'];
      if (!notify) {
        return;
      }
      // get the current tab, so we only create 1 notification per tab
      browser.tabs.query({ active: true, currentWindow: true }, (tabs) => {
        if (!tabs) {
          return;
        }
        if (tabs.length === 0) {
          return; // For example: only the background devtools or a popup are opened
        }

        const tab = tabs[0];
        if (sessionStorage.getItem(`malwareNotification${tab.id}`)) {
          // we've already notified the user, just return.
          return;
        }
        sessionStorage.setItem(`malwareNotification${tab.id}`, true);

        const notificationOptions = {
          title: 'AdBlock',
          iconUrl: browser.extension.getURL('img/icon48.png'),
          type: 'basic',
          priority: 2,
          message: translate('malwarenotificationmessage'),
        };
        // Pop up a notification to the user.
        browser.notifications.create(
          (Math.floor(Math.random() * 3000)).toString(),
          notificationOptions,
          () => {
          // do nothing in callback
          },
        );
      });// end of browser.tabs.query
    }); // end of browser.storage.local.get
  }// end of if
};// end of createMalwareNotification function

// Attach methods to window
Object.assign(window, {
  adblockIsPaused,
  addCustomFilter,
  addToMatchCache,
  confirmRemovalOfCustomFiltersOnHost,
  createPageWhitelistFilter,
  createWhitelistFilterForYoutubeChannel,
  emitPageBroadcast,
  getCustomFiltersText,
  getExcludeFiltersText,
  getFrameData,
  getSubscriptions,
  getSubscriptionsMinusText,
  getAdblockUserId,
  getDebugInfo,
  getFirstRun,
  getMalwareDomains,
  handlerBehaviorChanged,
  launchResourceblocker,
  openTab,
  pageIsWhitelisted,
  processFrameData,
  resetMatchCache,
  setCustomFiltersText,
  setExcludeFilters,
  setFirstRunToFalse,
  subscribe,
  tryToUnwhitelist,
  unsubscribe,
  updateSubscriptionsNow,
  updateButtonUIAndContextMenus,
  updateCustomFilterCountMap,
  validateFilter,
});

portInit.then(initializeBackgroundObjects);
log('done loading');
